import { EventEmitter } from 'events';
import { IServiceType, IHapEvInstance } from './interfaces';
import { createConnection, HttpMessageParser } from './eventedHttpClient';
import { CharacteristicType } from './hapClient';
import { log } from '../Types/types';

export class HapMonitor extends EventEmitter {
    private pin: string;
    private evInstances: IHapEvInstance[];
    private services: IServiceType[];
    private logger: log;
    private debug: log;
    private parser: HttpMessageParser

    constructor(logger: log, debug: log, pin: string, services: IServiceType[]) {
        super();
        this.logger = logger;
        this.debug = debug;
        this.pin = pin;
        this.services = services;
        this.evInstances = [];
        this.parser = new HttpMessageParser();

        // get a list of characteristics we can watch for each instance
        this.parseServices();

        // start watching
        this.start();
    }

    /**
     * Start monitoring
     */
    public start(): void {
        for (const instance of this.evInstances) {
            instance.socket = createConnection(instance, this.pin, { characteristics: instance.evCharacteristics });

            this.debug(`[HapClient] [${instance.ipAddress}:${instance.port} (${instance.username})] Connected`);

            instance.socket.on('data', (data) => {
                const message = this.parser.parse(data);


                if (message.statusCode === 401) {
                    if (this.logger) {
                        this.logger(`[HapClient] [${instance.ipAddress}:${instance.port} (${instance.username})] ` +
                            `${message.statusCode} ${message.statusMessage} - make sure Homebridge pin for this instance is set to ${this.pin}.`);
                    }
                }

                if (message.protocol === 'EVENT') {
                    try {
                        const body = JSON.parse(message.body);
                        if (body.characteristics && body.characteristics.length) {
                            this.debug(`[HapClient] [${instance.ipAddress}:${instance.port} (${instance.username})] ` +
                                `Got Event: ${JSON.stringify(body.characteristics)}`);

                            const response = body.characteristics.map((c: CharacteristicType): IServiceType | undefined => {
                                // find the matching service for each characteristics
                                const service = this.services.find(x => x.aid === c.aid && x.instance.username === instance.username);

                                if (service) {
                                    // find the correct characteristic and update it
                                    const characteristic = service.serviceCharacteristics.find(x => x.iid === c.iid);
                                    if (characteristic) {
                                        characteristic.value = c.value;
                                        service.values[characteristic.type] = c.value;
                                        return service;
                                    }
                                }

                                return undefined;

                            });

                            // push update to listeners
                            this.emit('service-update', response.filter((x: unknown) => x));
                        }
                    } catch (e) {
                        // do nothing
                    }
                }
            });
        }
    }

    /**
     * Stop monitoring.
     */
    public stop(): void {
        for (const instance of this.evInstances) {
            if (instance.socket) {
                try {
                    instance.socket.destroy();
                    this.debug(`[HapClient] [${instance.ipAddress}:${instance.port} (${instance.username})] Disconnected`);
                } catch (e) {
                    // do nothing
                }
            }
        }
    }

    private parseServices(): void {
        // get a list of characteristics we can watch for each instance
        for (const service of this.services) {
            const evCharacteristics = service.serviceCharacteristics.filter(x => x.perms.includes('ev'));

            if (evCharacteristics.length) {
                // register the instance if it's not already there
                if (!this.evInstances.find(x => x.username === service.instance.username)) {
                    const newInstance = Object.assign({}, service.instance) as IHapEvInstance;
                    newInstance.evCharacteristics = [];
                    this.evInstances.push(newInstance);
                }

                const instance = this.evInstances.find(x => x.username === service.instance.username);

                for (const evCharacteristic of evCharacteristics) {
                    if (instance &&
                        instance.evCharacteristics &&
                        !instance.evCharacteristics.find(x => x.aid === service.aid && x.iid === evCharacteristic.iid)) {
                        instance.evCharacteristics.push({ aid: service.aid, iid: evCharacteristic.iid, ev: true });
                    }
                }
            }
        }
    }
}
